/*****************************************************************************************
 *
 * Copyright (C), 2005
 * SAS Institute Inc., Cary, N.C. 27513, U.S.A.  All rights reserved.
 * 
 * NAME:    orvinc_fitting                                      
 *       
 * PURPOSE: Wrapper for orva_sfitall and orva_sfit_kri macros that gets all
 *          necessary fitting options out of the fitoptsData or initData
 *          for severity fitting and them appends new fits to outparm dataset.
 *
 * SUPPORT: jariou
 *
 * PARAMETERS:
 *     data             - loss data
 *     datatype         - datatype I(nternal), (C)onsortium, (E)xternal
 *     ccUid            - ccUid for cell to fit
 *     ccData           - Cross Classification dataset
 *     useCellOverrides - use cell overrides (vs project options)
 *     initData         - sparm dataset to use as starting point
 *                        for fitting during incremental load
 *     iteration        - iteration to refit (only used with initData)
 *     fitoptsData      - fitting options (NA if 'initData' is passed)
 *     sparmMap         - dataset with sparm datasets to use for probability of observability
 *     outparm          - out sparm dataset to write results to
 *     outcov           - out sparm covariance dataset to write results to
 *     parallel         - parallel processing?
 *
 * NOTES:  None
 *****************************************************************************************/
%macro orva_sfit( 
                 data             = ,
                 datatype         = , 
                 ccUid            = ,          
                 ccData           = ,    
                 useCellOverrides = false,      
                 initData         = , 
                 fitoptsData      = ,
                 sparmMap         = ,     
                 outparm          = ,
                 outcov           = ,
                 parallel         = 0,
                 iteration        = ,
                 CTool            = false
                 );
   %if (%upcase(&ORV_DEBUG.) EQ TRUE) %then %do;            
      %put #orva_sfit(;
      %put .           data             = &data.,;
      %put .           datatype         = &datatype.,;
      %put .           ccUid            = &ccUid.,;      
      %put .           ccData           = &ccData.,; 
      %put .           useCellOverrides = &useCellOverrides.,; 
      %put .           initData         = &initData.,; 
      %put .           fitoptsData      = &fitoptsData.,;
      %put .           sparmMap         = &sparmMap.,;     
      %put .           outparm          = &outparm.,;
      %put .           outcov           = &outcov.,;
      %put .           parallel         = &parallel.,;
      %put .           iteration        = &iteration.,;
      %put .           CTool            = &CTool.;
      %put .           );
   %end;

   %let orva_sfit_abort  = 0;
   %let orva_sfit_skip   = 0;
   %let orva_sfit_reason = %sysfunc(sasmsg(&orvads, unknown_error, noquote));
   
   /*- initialize Probability of Observability macro variables -*/
   %let Probability_Observed = .;
   %let Source_Method        = .;
   %let Source_Type          = .;
   %let Source_Uid           = 0;
   %let Use_Kri_Owners       = false;
   %let Kri_Owners           = %str();
   
   %if &Iteration ne %str() %then %do;
      %let refit = true;
   %end;
   %else %do;
      %let refit = false;
   %end;
   
   /***************************************************************/
   /*- For incremental load, all options are in initData dataset -*/
   /***************************************************************/
   %if &initData NE %str() %then %do;
      /* default values */
      %let SevUserInput        = false;
      %let TailBoundaryPercent = 95;
      %let TailBoundaryValue   = 1000000;
      %let TailBoundaryChoice  = TailBoundaryPercent;
      
      data _NULL_; 
         length KRI_LOG $2000;
         set &initData; 
         where 
               CCUID = &CCUID
               %if &iteration NE %str() %then %do;
                  and ITERATION eq &iteration
               %end;
               ;

         /*- get KRI Controls -*/
         %if &datatype EQ I %then %do;
            if COUNT_KRI > 0 then do;
               call symput('Sev_Use_Kris', 'true');       
            end;
            else do;
               call symput('Sev_Use_Kris', 'false');  
            end;
         %end;
         %else %do;
            call symput('Sev_Use_Kris', 'false');
         %end;

         call symput('Sev_Selection_Method_Kris', 'Stepwise');
         
         /*- get Threshold options -*/
         call symputx('threshold',    threshold);
         call symputx('gofThreshold', gofThreshold);

         if defaultThreshold eq . then do;
            call symput('defaultThreshold', 0);
         end;
         else do;
            call symputx('defaultThreshold', put(defaultThreshold, BEST20.));
         end;
         
         call symput("Sev_Keep_Larger_Thresholds", Sev_Keep_Larger_Thresholds);

         if Sev_Keep_Larger_Thresholds eq . then do;
            call symput("Sev_Keep_Larger_Thresholds", "FALSE");
         end;
 
         /*- get Probability of Observability options -*/
         if percentThreshold eq . then do;
            call symput('thresholdPercent', -1);
         end;
         else do;
            call symput('thresholdPercent', put(percentThreshold, BEST20.));
         end;

         /*- get Distribution -*/
         call symput("dists", Distribution);

         %do i = 1 %to &MAX_SEV_PARAMS;
            if ^missing(Param_&i) then do;
              call symputx("Param_&i._init", put(Param_&i., BEST20.));
            end;
         %end;

         call symputx("nParms", paramCount);
         call symputx("Min_Fitting_Count", Min_Fitting_Count);
         
         /*- get SevUserInput options -*/
         call symputx("param_xlimit", param_xlimit);

         if param_xlimit ne . then do;
           call symputx("param_pnlimit", param_pnlimit);
         end;
         
         /*- get KRIs -*/
         call symput('Sev_Kris_Forced',    '');          
         call symput('Sev_Kris_Suggested', '');      

         if Use_Kri_Owners EQ '' then do;
            call symput('Use_Kri_Owners', 'false');
         end;
         else do;
            call symput('Use_Kri_Owners', Use_Kri_Owners);
         end;
         
         %if &datatype EQ I %then %do;
            array KRI Param_1_:;
            length KRI_LIST $32000;

            if COUNT_KRI GT 0 then do;
              do i = 1 to dim(KRI);
                  if KRI[i] NE . then do;
                     if substrn(vname(KRI(i)),9,4) = "KRI_" then do;
                         KRI_LIST = trim(KRI_LIST) || "  " || trim(substrn(vname(KRI(i)), 9)) ;
                     end;
                  end;
              end;

              call symput(  "Sev_Kris_Forced"   , trim(left(KRI_LIST)));   
              call symput(  "KRI_LOG"           , trim(left(KRI_LOG)));
              call symputx( "KRI_EFFECT"        , KRI_EFFECT  );
              call symputx( "KRI_EFFECT_RETAIN" , KRI_EFFECT_RETAIN );
              call symputx( "KRI_LOSS_EFFECT"   , KRI_LOSS_EFFECT  );
              call symputx( "KRI_RETAIN"        , KRI_RETAIN  );
              call symputx( "KRI_AVG_EFFECT"    , KRI_AVG_EFFECT  );
            end;
        %end;
      run;     
   %end;
 
   /******************************************************/
   /*- Otherwise, get options from fitopts dataset      -*/
   /******************************************************/
   %else %if &fitoptsData NE %str() %then %do;
      %if &useCellOverrides EQ false %then %do;
         %let ccuidForOptions = .;
      %end;
      %else %do;
         %let ccuidForOptions = &ccUid;
      %end;

      /*- get Distribution options -*/
      %let dists = ;
      %orva_readFitOpts(
                        fitoptsdata = &fitoptsData, 
                        type        = DIST, 
                        fittype     = SEV,
                        ccUid       = &ccuidForOptions,
                        datatype    = &datatype
                        );
      %if &dists EQ %str() %then %do;
         %put &orvar_error %sysfunc(sasmsg(&orvads, sfit_selected_error, noquote));
         %let orva_sfit_reason = %sysfunc(sasmsg(&orvads, sfit_selected_error, noquote));
         %let orva_sfit_abort  = 1;
         %goto DONE;
      %end;  
      
      %let Min_Fitting_Count = ;
      %orva_readFitOpts(
                        fitoptsdata = &fitoptsData,
                        type        = ADVANCED,
                        fittype     = SEV,
                        ccUid       = &ccuidForOptions,
                        datatype    = &datatype
                        );
      %if &Min_Fitting_Count EQ %str() %then %do;
         %let orva_sfit_reason = %sysfunc(sasmsg(&orvads, sfit_no_minimmum_data_req_error, noquote));
         %put &orvar_error &orva_sfit_reason;
         %let orva_sfit_abort = 1;
         %goto DONE;
      %end;  
      
      /*- get KRI Control Options -*/
      %let Sev_Use_Kris              = false;
      %let Sev_Selection_Method_Kris = Stepwise;

      %if &datatype EQ I %then %do;
         %orva_readFitOpts(
                           fitoptsdata = &fitoptsData,
                           type        = KRICONTROL,
                           fittype     = SEV,
                           ccUid       = &ccuidForOptions,
                           datatype    = &datatype
                           );
      %end;

      /*- get Threshold options -*/
      %let threshold                  = . ;
      %let gofThreshold               = . ;
      %let defaultThreshold           = . ;
      %let Sev_Keep_Larger_Thresholds = . ;
      %orva_readFitOpts(
                        fitoptsdata = &fitoptsData,
                        fittype     = SEV,   
                        type        = THRESHOLD,
                        datatype    = &datatype,
                        ccUid       = &ccuidForOptions
                        );          

      /*- get probability of Observability -*/
      %orva_read_pobserv_opts(
                              fittype          = SEV,
                              datatype         = &datatype,
                              ccUid            = &ccuidForOptions,
                              fitoptsData      = &fitoptsData,
                              sparmMap         = &sparmMap,
                              lossData         = &data,
                              threshold        = &threshold,
                              defaultThreshold = &defaultThreshold
                              );   

      %if &Probability_Observed eq . %then %do;
        %let thresholdPercent = -1;
      %end;
      %else %do;
        %let thresholdPercent = %sysevalf(1 - &Probability_Observed);
      %end; 
      
      /*- get SevUserInput options -*/
      %let TailBoundaryPercent = . ;
      %let TailBoundaryValue   = . ;
      %let TailBoundaryChoice  = . ;
      %let SevUserInput        = false;
      %orva_readFitOpts(
                        fitoptsdata = &fitoptsData,
                        type        = SEVUSERINPUT,
                        ccUid       = &ccuidForOptions,
                        datatype    = &datatype
                        );
      
      /*- get KRIs -*/
      %let Sev_Kris_Suggested = ;
      %let Sev_Kris_Forced    = ;

      %if &Sev_Use_Kris EQ true %then %do;
         %orva_readFitOpts(
                           fitoptsdata = &fitoptsData,
                           type        = KRI,
                           fittype     = SEV,
                           ccUid       = &ccuidForOptions
                           );
         %if &Sev_Kris_Suggested EQ %str() AND &Sev_Kris_Forced EQ %str() %then %do;
            %let Sev_Use_Kris = false;
         %end;       
      %end;
   %end;

   /***************************************************/
   /*** always get some options from fitoptsDataset ***/
   /***************************************************/
   %let Sev_FitMeasures   = LOGL 1 KS_V 0 SBC 0 AIC 0 CVM_V 0 AD_V 0;
   %let Sev_FavFitMeasure = LOGL;

   %if &fitoptsData NE %str() %then %do;
      %orva_readFitOpts(
                        fitoptsdata = &fitoptsData,
                        type        = FITMEASURE,
                        ccUid       = &ccuidForOptions,
                        datatype    = &datatype
                        );
   %end;
  
   /*- create subsetting where for loss dataset -*/
   %let NUMCCVS    = 0;
   %let CELL_WHERE = ;
   %orvu_cell_where(
                    ccuid  = &ccuid, 
                    ccData = &ccData
                    );

   %if &NUMCCVS eq 0 %then %do;
      %put &orvar_error %sysfunc(sasmsg(&orvads, sfit_subset_error, noquote, &ccUid, &data));
      %let reason = %sysfunc(sasmsg(&orvads, sfit_subset_error, noquote, &ccUid, &data));
      %let abort  = 1;
      %goto DONE;
   %end;
   %else %do;
      %let abort  = 0;
      %let status = dists;
      %let reason = %sysfunc(sasmsg(&orvads, successful_note, noquote));
   %end;
   
   /*****************************************************************
    * When in the CTool create indicator variables for owners,
    * unless we're doing a refit and there aren't any Kri_Owners 
    * stored.           
    ****************************************************************/
   %let Sev_Use_Kris_Dummy = false;
   %let ownerCount = 1;

   %if &CTool eq true and ((&refit eq false) or (&Use_Kri_Owners eq true)) %then %do;
      %let KRI_Owners = %str();
      %orva_create_own_dummy( 
                             inData      = &data,
                             whereClause = &CELL_WHERE,
                             outData     = _sfit_owner_data
                             );
   %end;
   
   /*******************************************************
    * Perform regular fit, if we're not doing a refit or
    * we are refittinng the non-owner iteration                      
    *******************************************************/
   %if &refit EQ false OR &Sev_Use_Kris_Dummy EQ false %then %do;
      %if &Sev_Use_Kris EQ true %then %do;
        %put &orvar_status %sysfunc(sasmsg(&orvads, sfit_call_sevparam_note, noquote, &Sev_Kris_Forced, &Sev_Kris_Suggested));
        %orva_sfit_kri( 
                       data        = &data,
                       whereClause = &CELL_WHERE,
                       outparm     = _sfit_tmpsparm,
                       outcov      = _sfit_tmpscov,
                       parallel    = &parallel
                       );
      %end;
      %else %do;
        %put &orvar_status %sysfunc(sasmsg(&orvads, sfit_all_call_note, noquote));
        %orva_sfitall(
                      data        = &data,
                      whereClause = &CELL_WHERE,
                      outparm     = _sfit_tmpsparm,
                      outcov      = _sfit_tmpscov,
                      parallel    = &parallel
                      );
      %end; 
   
      %let Use_KRI_Owners = false;
      %orv_finalizeIteration;
   %end;

   /************************************************
    * Perform CTool Owners fit
    ***********************************************/
   %if &Sev_Use_Kris_Dummy EQ true %then %do;
      %let &Sev_Use_Kris = true;
      %let Sev_Kris_Suggested = &Sev_Kris_Suggested &KRI_Owners;
      %put &orvar_status %sysfunc(sasmsg(&orvads, sfit_call_sevparam_note, noquote, &Sev_Kris_Forced, &Sev_Kris_Suggested));
      %orva_sfit_kri( 
                     data        = _sfit_owner_data,
                     whereClause = &CELL_WHERE,
                     useQLIM     = true,
                     outparm     = _sfit_tmpsparm,
                     outcov      = _sfit_tmpscov,
                     parallel    = &parallel
                     );
      
      %if "&KRI_Owners" eq "" %then %do;
         %let Use_KRI_Owners = false;
      %end;
      %else %do;
         %let Use_KRI_Owners = true;
      %end;

      %orv_finalizeIteration;
  %end;
/**********************************************/
/***** end of macro ***************************/
/**********************************************/
%DONE:
   
   %if &orva_sfit_abort %then %do;
      %put &orvar_error %sysfunc(sasmsg(&orvads, sfit_abnormal_error, noquote));
   %end;

   %if &orva_sfit_skip %then %do;
      %put &orvar_error &reason;
   %end;  

   %orvu_delete_data( data = _sfit_: );
%mend orva_sfit;

%macro orv_finalizeIteration;
   /*******************************************************************/
   /**** add Fit Options to parm dataset that may not be there    *****/
   /*******************************************************************/
   data _sfit_tmpsparm; 
      set _sfit_tmpsparm;
      length KRI_FORCED     $32000;
      length KRI_SUGGESTED  $32000;
      length POBSERV_STYPE  $1; 
      length Use_Kri_Owners $8;
      length POBSERV_METHOD $32;

      POBSERV_METHOD    = "&Source_Method";
      MIN_FITTING_COUNT = &Min_Fitting_Count;

      %if &Source_Type NE . %then %do;
         POBSERV_STYPE = "&Source_Type";
      %end;
      %else %do; 
         POBSERV_STYPE = "";
      %end;

      POBSERV_SUID = &Source_Uid;

      %if &datatype EQ I %then %do;
        KRI_FORCED     = "&Sev_Kris_Forced";
        KRI_SUGGESTED  = "&Sev_Kris_Suggested";
        Use_Kri_Owners = "&Use_Kri_Owners";
      %end;
      %else %do;
        KRI_FORCED     = "";
        KRI_SUGGESTED  = "";
        Use_Kri_Owners = 'false';
      %end;

      /* No Need for param_plimit in here... */
      drop param_plimit;
   run;

   %let nobs = 0;
   %let dsid = %sysfunc(open( _sfit_tmpscov, I));

   %if &dsid ^= 0 %then %do;
     %let nobs = %sysfunc(attrn(&dsid,nlobs));
     %let rc   = %sysfunc(close(&dsid));
   %end;

   %if &nobs ne 0 %then %do;
      data _sfit_tmpscov; 
         set _sfit_tmpscov;
         length _NAME_ $ 32;

         if rowname ne '' then  do;
            _NAME_ = rowname;
         end;
         else do;
            _NAME_ = variable;
         end;

         drop rowname variable;
      run;
   %end;
   
   /**********************************************/
   /***** append new fits to outparm dataset *****/
   /**********************************************/
   %orva_append_parm_dataset(
                             tmpParm     = _sfit_tmpsparm,
                             finalParm    = &outparm,
                             tmpCov       = _sfit_tmpscov,
                             finalCov     = &outcov,
                             ccUid        = &ccUid,
                             measCol      = &Sev_FitMeasures,
                             selectedMeas = &Sev_FavFitMeasure
                             %if &iteration NE %str() %then %do;
                             , iteration    = %eval(&iteration - 1)
                             %end;
                             );
%mend orv_finalizeIteration;
