/********************************************************************
 * Copyright (C), 2004
 * SAS Institute Inc., Cary, N.C. 27513, U.S.A.  All rights reserved.
 * 
 *   Title: orvd_LNG.sas
 *   
 *   -Provides support for the LogNormal-Gamma distribution-
 *
 *        This file contains macro fragments to implement calculation specific to the 
 *        LogNormal-Gamma distribution
 *
 *  The macro %ORVD_LNG is at the end of this file. It is an empty macro that takes
 *  no parameter and returns no feedback. In normal usage, this macro needs to be 
 *  called before trying to call any other macros depending on the fragments in this file.
 *
 *  Calling the %ORVD_LNG macro makes the SAS system parse and compile the whole file 
 *  so that the other macro fragment then become defined and available for other 
 *  macros to call.
 *
 *  Of course this is only guaranteed to work if there is no syntax error in any of 
 *  the fragments. 
 *
 ********************************************************************/
%macro LNG_rep(q = q);
   do _lng_rep_i = 1 to &LNG_NQUANT;
        if Param_3 < 0.00002 then do;
            &q[_lng_rep_i] = sqrt(1 + probit(_lng_rep_i / (&LNG_NQUANT + 1)) * sqrt(Param_3));
        end;
        else do;        
            &q[_lng_rep_i] = sqrt(gaminv(_lng_rep_i / (&LNG_NQUANT + 1), 1 / Param_3) * Param_3);
        end;
   end;
%mend LNG_rep;   

/*****************************************************************************************
 *   What it does  :  Evaluate the cumulative distribution function (CDF)
 *
 *   Parameters    :
 *              x     = value at which the CDF is evaluated
 *        returnVal = name of variable inton which result is stored
 *****************************************************************************************/
%macro CdfLNG(x, returnVal);
    if &Param_3 < 1e-8 then do;
        &returnVal = cdf("LOGNORMAL", &x, &Param_1, &Param_2);   
    end;
    else do;
        &returnVal = 0;
        retain q&returnVal._1-q&returnVal._&LNG_NQUANT;
        array q&returnVal q&returnVal._1-q&returnVal._&LNG_NQUANT; 

        if _N_ = 1 then do;
            do i_lng_rep = 1 to &LNG_NQUANT;
                if &Param_3 < 0.00002 then do;
                    q&returnVal[i_lng_rep] = sqrt(1 + probit(i_lng_rep / (&LNG_NQUANT + 1)) * sqrt(&Param_3));
                end;
                else do;
                    q&returnVal[i_lng_rep] = sqrt(gaminv(i_lng_rep / (&LNG_NQUANT + 1), 1 / &Param_3) * &Param_3);
                end;
            end;
        end; 

        do i_lng_rep = 1 to &LNG_NQUANT;
            if &Param_2 * q&returnVal[i_lng_rep] < 1e-8 then do;
                &returnVal = &returnVal + 1;
            end;
            else do;
                &returnVal = &returnVal + cdf('LOGNORMAL', &x, &Param_1, &Param_2 * q&returnVal[i_lng_rep]);
            end; 
        end;

        &returnVal = &returnVal / &LNG_NQUANT;
        drop q&returnVal._1-q&returnVal._&LNG_NQUANT i_lng_rep;
    end;
%mend CdfLNG;

%macro VCdfLNG(x, returnVal);
    if Param_3 < 1e-8 then do;
        &returnVal = cdf("LOGNORMAL",(&x), Param_1, Param_2);   
    end;
    else do;
        &returnVal = 0;
        retain qa&returnVal._1-qa&returnVal._&LNG_NQUANT;
        array qa&returnVal qa&returnVal._1-qa&returnVal._&LNG_NQUANT; 

        if (_N_ = 1 ) or missing(_N_) then do;
            do i_lng_rep = 1 to &LNG_NQUANT;
                if Param_3 < 0.00002 then do;
                    qa&returnVal[i_lng_rep] = sqrt(1 + probit(i_lng_rep / (&LNG_NQUANT + 1)) * sqrt(Param_3));
                end;
                else do;
                    qa&returnVal[i_lng_rep] = sqrt(gaminv(i_lng_rep / (&LNG_NQUANT + 1), 1 / Param_3) * Param_3);
                end;
            end;
        end; 

        do i_lng_rep = 1 to &LNG_NQUANT;
            if Param_2 * qa&returnVal[i_lng_rep] < 1e-8 then do;
                &returnVal = &returnVal + 1;
            end;
            else do;
                &returnVal = &returnVal + cdf('LOGNORMAL', (&x), Param_1, Param_2 * qa&returnVal[i_lng_rep]);
            end; 
        end;

        &returnVal = &returnVal / &LNG_NQUANT;
        drop qa&returnVal._1-qa&returnVal._&LNG_NQUANT i_lng_rep;
    end;
%mend VCdfLNG;

/*****************************************************************************************
 *   What it does  :  Evaluate the density distribution function (CDF)
 *
 *   Parameters    :
 *              x     = value at which the density is evaluated
 *        returnVal = name of variable inton which result is stored
 *****************************************************************************************/
%macro PdfLNG(x, returnVal);
   if &Param_3 < 1e-8 then do;
          &returnVal = pdf("LOGNORMAL", &x, &Param_1, &Param_2);   
   end;
   else do;
          &returnVal = 0;
          retain q1-q&LNG_NQUANT;
          array q q1-q&LNG_NQUANT; 

          if _N_ = 1 then do;
              do i_lng_rep = 1 to &LNG_NQUANT;
                  if &Param_3 < 0.00002 then do;
                          q[i_lng_rep] = sqrt(1 + probit(i_lng_rep / (&LNG_NQUANT + 1)) * sqrt(&Param_3));
                  end;
                  else do;
                          q[i_lng_rep] = sqrt(gaminv(i_lng_rep / (&LNG_NQUANT + 1), 1 / &Param_3) * &Param_3);
                  end;
              end;
          end; 
                  
          do i_lng_rep = 1 to &LNG_NQUANT;
                  &returnVal = &returnVal + pdf('LOGNORMAL', &x, &Param_1, &Param_2 * q[i_lng_rep]);
          end; 

          &returnVal = &returnVal / &LNG_NQUANT;
          drop q1-q&LNG_NQUANT i_lng_rep;
   end;
%mend PdfLNG;

%macro VPdfLNG(x, returnVal);
   if Param_3 < 1e-8 then do;
          &returnVal = pdf("LOGNORMAL", (&x), Param_1, Param_2);   
   end;
   else do;
          &returnVal = 0;
          retain qb1-qb&LNG_NQUANT;
          array qb qb1-qb&LNG_NQUANT; 

          if (_N_ = 1 ) or missing(_N_) then do;
              do i_lng_rep = 1 to &LNG_NQUANT;
                  if Param_3 < 0.00002 then do;
                          qb[i_lng_rep] = sqrt(1 + probit(i_lng_rep / (&LNG_NQUANT + 1)) * sqrt(Param_3));
                  end;
                  else do;
                          qb[i_lng_rep] = sqrt(gaminv(i_lng_rep / (&LNG_NQUANT + 1), 1 / (Param_3)) * Param_3);
                  end;
              end;
          end; 
                  
          do i_lng_rep = 1 to &LNG_NQUANT;
                  &returnVal = &returnVal + pdf('LOGNORMAL', &x, Param_1, Param_2 * qb[i_lng_rep]);
          end; 

          &returnVal = &returnVal / &LNG_NQUANT;
          drop qb1-qb&LNG_NQUANT i_lng_rep;
   end;
%mend VPdfLNG;

/*****************************************************************************************
 *   What it does  :  Generates a dataset named _temp_ containing probability values
 *                    and their corresponding quantiles
 *
 *   No Parameters    
 *****************************************************************************************/
%macro QuantLNG(scaleFactor = 1);
    %orvd_LOGN;
    %QuantLOGN(scaleFactor = &scaleFactor);
    %orvd_LNG;

    %if %sysevalf( &Param_3 > 1.0e-8, boolean) %then %do;
     
        data _parm_;
            mu    = &Param_1;
            sigma = &Param_2;
            beta  = &Param_3;
        run;

        data _temp_;
            set _temp_; 
            quant = log(quant);
        run;

        proc model 
                  data      = _temp_ 
                  parmsdata = _parm_ 
                  out       = _temp1_ 
                  noprint
                  ;
            parms 
                  mu 
                  sigma 
                  beta
                  ;
            F = 0;
            T = 0;

            do i_lng_rep = 1 to &LNG_NQUANT;
                if beta < 0.00002 then do;
                    q = sqrt(1 + probit(i_lng_rep / (&LNG_NQUANT + 1)) * sqrt(beta));
                end;
                else do;
                    q = sqrt(gaminv(i_lng_rep / (&LNG_NQUANT + 1), 1 / beta) * beta);;
                end;

                F = F + probnorm((quant - mu) / sigma / q) / &LNG_NQUANT;

                if THRESHOLD NE 0 then do;
                    T = T + probnorm((log(THRESHOLD) - mu) / sigma / q) / &LNG_NQUANT;
                end;
            end;

            F = (F - T * (&scaleFactor)) / (1 - T * (&scaleFactor)) / (&scaleFactor);
            eq.one = F - ecdf;        
            solve quant / converge = 1e-6;
        run;
        quit;   

        data _temp_;
            merge 
                  _temp_ 
                  _temp1_(keep = ecdf quant)
                  ;
            quant = exp(quant);
        run;
    %end;
%mend QuantLNG;

/*****************************************************************************************
 *   What it does  :  Calculates derivatives of the CDF function wrt parameters
 *
 *   Parameters    :
 *        scaleFactor = possible scaling for mixed distribution
 *        indata      = dataset of quantile points at which derivative is evaluated
 *        outdata     = dataset with derivatives 
 *****************************************************************************************/
%macro CdfDerLNG(
                 scaleFactor = 1, 
                 indata      = , 
                 outdata     =
                 );
   data _temp_0;
      set &indata;

      %if %sysevalf(&Param_3 < 1e-8, boolean) %then %do;
         cdf=cdf("LOGNORMAL", quant, &Param_1, &Param_2);   
      %end;
      %else %do;
         cdf = 0;
         retain qcdf_1-qcdf_&LNG_NQUANT;
         array qcdf qcdf_1-qcdf_&LNG_NQUANT; 

         if _N_ = 1 then do;
            do i_lng_rep = 1 to &LNG_NQUANT;
               if &Param_3 < 0.00002 then do;
                  qcdf[i_lng_rep] = sqrt(1 + probit(i_lng_rep / (&LNG_NQUANT + 1)) * sqrt(&Param_3));
               end;
               else do;
                  qcdf[i_lng_rep] = sqrt(gaminv(i_lng_rep /(&LNG_NQUANT + 1), 1 / &Param_3) * &Param_3);
               end;
            end;
         end; 

         do i_lng_rep = 1 to &LNG_NQUANT;
            if &Param_2 * qcdf[i_lng_rep] < 1e-8 then do;
               cdf = cdf + 1;
            end;
            else do;
               cdf = cdf + cdf('LOGNORMAL', quant, &Param_1, &Param_2 * qcdf[i_lng_rep]);
            end; 
         end;

         cdf  = &scalefactor * cdf / &LNG_NQUANT;
         cdf1 = cdf;
         drop i_lng_rep;
      %end;
   run;

   proc model data = _temp_0 noprint;
      %if %sysevalf(&Param_3 < 1e-8, boolean) %then %do;
         dependent cdf;
         retain Param_1 &Param_1 Param_2 &Param_2;
         cdf = cdf("LOGNORMAL", quant, Param_1, Param_2);   

         DParam_2 = getder(pred.cdf, Param_2);
         DParam_3 = 0;

         outvars 
                  DParam_2 
                  DParam_3 
                  quant
                  ;
         solve Param_2 /out = &outdata(keep = 
                                               quant 
                                               cdf 
                                               DParam_2 
                                               DParam_3
                                               );
      %end;
      %else %do;
         dependent cdf cdf1 ;
         retain 
                Param_1 &Param_1 
                Param_2 &Param_2 
                Param_3 &Param_3
                ;
         temp_cdf = 0;

         do i_lng_rep = 1 to &LNG_NQUANT;
            if Param_3 < 0.00002 then do;
               if Param_2*sqrt(1 + probit(i_lng_rep/(&LNG_NQUANT+1)) * sqrt(Param_3)) < 1e-8 then do;
                  temp_cdf = temp_cdf + 1;
               end;
               else do;
                  temp_cdf = temp_cdf + cdf('LOGNORMAL',quant,Param_1,Param_2*sqrt(1 + probit(i_lng_rep/(&LNG_NQUANT+1)) * sqrt(Param_3)));
               end; 
            end;
            else do;
               if (Param_2*sqrt(gaminv(i_lng_rep/(&LNG_NQUANT+1),1/Param_3) * Param_3) < 1e-8) then do;
                  temp_cdf = temp_cdf + 1;
               end;
               else do;
                  temp_cdf = temp_cdf + cdf('LOGNORMAL',quant,Param_1,Param_2*sqrt(gaminv(i_lng_rep/(&LNG_NQUANT+1),1/Param_3) * Param_3));
               end; 
            end;
         end;

         cdf  = &scalefactor*temp_cdf/&LNG_NQUANT;
         cdf1 = pred.cdf;

         DParam_2 = getder(pred.cdf,Param_2);
         DParam_3 = getder(pred.cdf1,Param_3);

         outvars 
                  DParam_2 
                  DParam_3 
                  quant
                  ;
         solve Param_2 Param_3 /out = &outdata(keep = 
                                                      quant 
                                                      cdf 
                                                      DParam_2 
                                                      DParam_3
                                                      );
      %end;
   quit;
%mend CdfDerLNG;

/*****************************************************************************************
 *   What it does  :  Simulates a random value
 *
 *   No Parameters    :  
 *****************************************************************************************/
%macro SimLNG(loss);
        /*        
        This fix is done here because when the "Param_3" parameter becomes too small,
        the simulation of the gamma variate fails. However this corresponds to a very
        peaky distribution around 1. Another way to look at this is that when the Param_3 
        parameter is so small, the distribution degenerates to a log-normal and that is
        the choice that should have been made if one had used the AIC or BSC instead of 
        simply the best likellihood.
        */
        if (Param_3 < 0.000000001) then do;
                /*         Take absolut value here because we need to take a square
                        root later. This introduces a minimal bias that can be
                        ignored because the probability of this ever occuring 
                        is extremely low */
                gamma = abs(rand("NORMAL") * sqrt(Param_3) + 1.0);
        end;
        else do;
                gamma = rand("GAMMA",1/Param_3) * Param_3;
        end;

        &loss = rand("NORMAL") * Param_2 * sqrt(gamma) + Param_1;
        &loss = exp(&loss);
%mend SimLNG;

/*****************************************************************************************
 *   What it does  :  Actually does the minimization using PROC NLP
 *
 *   No Parameters    
 *****************************************************************************************/
%macro InitFitLNG(
                  summaryData = , 
                  KRI_FIT     = NO
                  );
    %let nparms       = 3;
    %let Param_1_init = .;
    %let Param_2_init = .;
    %let Param_3_init = .;

    /*- initial guess based on moments -*/
    data _null_;
        set &summaryData; 
        mu    = sample_log_mean;        
        beta  = sample_log_kurtosis;
        sigma = sqrt(sample_log_variance);

        call symputx("Param_1_init", put(mu, BEST20.));
        call symputx("Param_2_init", put(sigma, BEST20.));
        /*call symputx("Param_3_init", put(beta, BEST20.));*/
        /* The original parameter_3 = beta doesn't converge in some cases.  Param_3=1.0 
           seems to converge all the time and get a better optimization comparing with 
           other alternative such as using beta as Param_3 and tech=nrr in proc nlp.
         */        
        call symputx("Param_3_init", put(1.0, BEST20.));         
        call symputx("obsCount", obsCount);
    run;
        
    %if %sysevalf(&Param_3_init < 0, boolean) %then %do;
        %let Param_3_init=0.01;
    %end;

    %if &dist NE MIXDIST %then %do;    
        %put **********************************;
        %put Distribution  =  LNG;
        %put Observations  =  &obsCount;
        %put Param_1_init  =  &Param_1_init;
        %put Param_2_init  =  &Param_2_init;
        %put Param_3_init  =  &Param_3_init;
        %put **********************************;
    %end;  

    %if %sysevalf(&obsCount > 5000, boolean) %then %do;
        %let sample_size=400;

        data _workdata7_; set _LogWorkdata_;
            if ( ranuni(348975) < &sample_size/&obsCount) then output;
        run;

        %FitLNG( inputdataset=_workdata7_ );

        data _NULL; set _parmnlp_(where=(_TYPE_="PARMS"));
            call symputx("Param_1_init", put(Param_1, BEST20.));
            call symputx("Param_2_init", put(Param_2, BEST20.));
            call symputx("Param_3_init", put(Param_3, BEST20.));        
        run;

        %if %sysevalf(&Param_3_init < 1.0e-2, boolean) %then %do;
            %let Param_3_init = 1.0e-2;
        %end;

        %if &dist NE MIXDIST %then %do;
            %put **********************************;
            %put Distribution    =  LNG;
            %put Observations    =  &obsCount;;
            %put Param_1_init    =  &Param_1_init;
            %put Param_2_init    =  &Param_2_init;
            %put Param_3_init    =  &Param_3_init;
            %put **********************************;
        %end;
    %end;
%mend InitFitLNG;
       
/*****************************************************************************************
 *   What it does  :  Actually does the minimization using PROC NLP
 *
 *   No Parameters    
 *****************************************************************************************/
%macro FitLNG( inputdataset = _LogWorkdata_);
   ods listing close;
   ods output 
             IterStop           =  _logl_
             ConvergenceStatus  =  _cs_ 
             covmat             = _covmat_ 
             ;        

   proc nlp 
            data      =  &inputdataset
            outest    =  _parmnlp_
            &PROC_NLP_OPTIONS.
            ;

      parms 
            Param_1 = &Param_1_init, 
            Param_2 = &Param_2_init, 
            Param_3 = &Param_3_init,
            ;

      bounds %Param_BoundsLNG();
      max logf;
      retain lng_varScale1-lng_varScale&lng_nquant;
      array lng_varScale lng_varScale1-lng_varScale&lng_nquant;

      _DELTA_ = 1.0e-6;
      if _obs_ = 1 then do;
         do lng_rep_i = 1 to &lng_nquant;
            if Param_3 < 0.00002 then do;
               lng_varScale[lng_rep_i] = sqrt(1 + probit(lng_rep_i/(&LNG_NQUANT+1)) * sqrt(Param_3));
            end;
            else do;        
               lng_varScale[lng_rep_i]=sqrt(gaminv(lng_rep_i/(&LNG_NQUANT+1),1/Param_3)*Param_3);
            end;
         end;
      end;

      lsdf     = 0;
      the_lsdf = 0;
      the_logf = 0;

      do lng_rep_i = 1 to &LNG_NQUANT;
         if &THRESHOLDPERCENT >= 0 then do;
            the_lsdf = the_lsdf + cdf("LOGNORMAL", THRESHOLD, Param_1 , Param_2 * lng_varScale[lng_rep_i]);
         end;
         else if THRESHOLD NE 0 then do;
            the_lsdf = the_lsdf + sdf("LOGNORMAL", THRESHOLD, Param_1, Param_2 * lng_varScale[lng_rep_i]);
         end;

         the_logf = the_logf + pdf("LOGNORMAL", AMOUNT, Param_1, Param_2 * lng_varScale[lng_rep_i]);
      end;

      if &THRESHOLDPERCENT >= 0 then do;
         lsdf = - &THRESHOLDPERCENT / (1 - &THRESHOLDPERCENT) * log( the_lsdf / &LNG_NQUANT + 1E-30);
      end;
      else if THRESHOLD NE 0 then do;
         lsdf = log(the_lsdf / &LNG_NQUANT + 1E-30);
      end;

      logf =    -lsdf + log(the_Logf / &LNG_NQUANT + 1E-30);
   quit;
   ods listing;
%mend FitLNG;

/*****************************************************************************************
 *   What it does  :  Actually does the minimization using PROC MODEL
 *
 *   No Parameters    
 *****************************************************************************************/
%macro FitKriLNG( inputdataset = _LogWorkdata_KRI_);
   ods listing close;
   ods output 
             IterStop           =  _logl_
             ConvergenceStatus  =  _cs_ 
             covmat             = _covmat_ 
             ;        

   proc nlp 
            data   =  &inputdataset.(drop=Param_1:)
            outest =  _parmnlp_
            &PROC_NLP_OPTIONS.
            ;

      parms 
            Param_0 = &Param_1_init, 
            Param_2 = &Param_2_init, 
            Param_3 = &Param_3_init,
            &KRI_INIT_PARM
            ;

      bounds %Param_BoundsLNG();

      max logf;

        retain lng_varScale1-lng_varScale&lng_nquant;
        array lng_varScale lng_varScale1-lng_varScale&lng_nquant;

        _DELTA_ = 1.0e-6;
        if _obs_=1 then do;
           do lng_rep_i = 1 to &lng_nquant;
                if Param_3 < 0.00002 then do;
                    lng_varScale[lng_rep_i] = sqrt(1 + probit(lng_rep_i/(&LNG_NQUANT+1)) * sqrt(Param_3));
                end;
                else do;        
                    lng_varScale[lng_rep_i]=sqrt(gaminv(lng_rep_i/(&LNG_NQUANT+1),1/Param_3)*Param_3);
                end;
           end;
         end;      

      Param_1 = Safe_ID(Param_0 + &KRI_MODEL_YES_INTERCEPT);

      lsdf     = 0;
      the_lsdf = 0;
      the_logf = 0;

      do lng_rep_i = 1 to &LNG_NQUANT;
         if &THRESHOLDPERCENT >= 0 then do;
            the_lsdf = the_lsdf + cdf("LOGNORMAL", THRESHOLD, Param_1 , Param_2 * lng_varScale[lng_rep_i]);
         end;
         else if THRESHOLD NE 0 then do;
            the_lsdf = the_lsdf + sdf("LOGNORMAL", THRESHOLD, Param_1, Param_2 * lng_varScale[lng_rep_i]);
         end;

         the_logf = the_logf + pdf("LOGNORMAL", AMOUNT, Param_1, Param_2 * lng_varScale[lng_rep_i]);
      end;

      if &THRESHOLDPERCENT >= 0 then do;
         lsdf = - &THRESHOLDPERCENT / (1 - &THRESHOLDPERCENT) * log(the_lsdf / &LNG_NQUANT + 1E-30);
      end;
      else if THRESHOLD NE 0 then do;
         lsdf = log(the_lsdf / &LNG_NQUANT + 1E-30);
      end;

      logf =    -lsdf + log(the_Logf / &LNG_NQUANT + 1E-30);
   quit;
   ods listing;
%mend FitKriLNG;


/*****************************************************************************************
 *   What it does  :  Calculates the moments
 *
 *   No Parameters 
 *****************************************************************************************/
/*- actually, the moments of logarithm of losses -*/
%macro MomentsLNG;
    %do i = 1 %to 4;
        if Param_3 * Param_2 * Param_2 < 2 / (&i * &i) then do;
            if Param_3 < 1.0e-8 then do;
                m&i = exp(&i * Param_1+ &i * &i * Param_2 * Param_2 / 2);
            end;
            else do;
               m&i = exp(&i * Param_1) * (1 - Param_3 * &i * &i * Param_2 * Param_2 / 2)**(-1 / Param_3);
            end;
        end;
        else do;
            m&i = .;
        end;
    %end;  

    MEAN     = m1;
    STD      = sqrt(m2 - m1 * m1);
    SKEWNESS = (m3 - 3 * m1 * m2 + 2 * (m1**3)) / (std**3);
    KURTOSIS = (m4 - 4 * m1 * m3 + 6 * (m1**2) * m2 - 3 * (m1**4)) / (std**4);
%mend MomentsLNG;

%macro LimMomentsLNG( 
                     i, 
                     x, 
                     returnVal
                     );
   array q_&i._ _lm_q_&i._1-_lm_q_&i._&lng_nquant;                
   %LNG_REP(q = q_&i._);
   &returnVal = 0;

   do _lm_quants = 1 to &LNG_NQUANT;
       _lm_tmp_Param_2 = Param_2 * q_&i._[_lm_quants];
      
       _lm_tmp = exp(&i * Param_1 + &i * &i * _lm_tmp_Param_2 * _lm_tmp_Param_2 / 2);
       _lm_tmp = _lm_tmp * cdf("NORMAL", (log(&x) - Param_1 - &i * _lm_tmp_Param_2 * _lm_tmp_Param_2)/_lm_tmp_Param_2);
       _lm_tmp = _lm_tmp + &x**&i * (1 - cdf("NORMAL", (log(&x) - Param_1) / _lm_tmp_Param_2));
      
      &returnVal = &returnVal + _lm_tmp;
   end;
   
   &returnVal = &returnVal / &lng_nquant;
%mend LimMomentsLNG;

/*****************************************************************************************
 *   What it does  :  Does method of moments estimation and assigns the corresponding 
 *                    values to the Param_1, Param_2, Param_3, and shape2 macros
 *
 *   No Parameters 
 *****************************************************************************************/
%macro MomFitLNG;
   %let Param_1 = .;
   %let Param_2 = .;
   %let Param_3 = .;        

   data _null_;
      p2 = sqrt(&variance);
      p3 = &kurtosis / 3 - 1;
      call symputx("Param_1", put(&mean, BEST20.));
      call symputx("Param_2", put(p2,    BEST20.));
      call symputx("Param_3", put(p3,    BEST20.));
   run;
%mend MomFitLNG;

%macro LNG_unboundedmoments;
   %do i = 1 %to 4;
      if Param_3 * Param_2 * Param_2 < 2 / (&i * &i) then do;
         m&i = (&i * Param_1) + log(1 - Param_3 * &i * &i * Param_2 * Param_2 / 2)*(-1 / Param_3); 
         m&i = exp(m&i); 
      end;
      else do;
         m&i = .;
      end;
   %end;  

   MEAN     = m1;
   STD      = sqrt(m2 - m1 * m1);
   SKEWNESS = (m3 - 3 * m1 * m2 + 2 * (m1**3)) / (std**3);
   KURTOSIS = (m4 - 4 * m1 * m3 + 6 * (m1**2) * m2 - 3 * (m1**4)) / (std**4);
%mend LNG_unboundedmoments;

%macro MomFit_real_LNG;
   %let Param_1_init = .;
   %let Param_2_init = .;
   %let Param_3_init = .;

   data _null_;
      c = &mean**2 + &variance;

      if c <= 0 then do; 
         p1 = .;
         p2 = .;
      end;
      else do;
         c = sqrt(c);
         p1 = log(&mean**2/c);
         p2 = sqrt(2*log(c/&mean));
      end;

      Param_3 = .5 /( 8 * p2**2);
      call symputx("Param_1_init", put(p1,      BEST20.));
      call symputx("Param_2_init", put(p2,      BEST20.));
      call symputx("Param_3_init", put(Param_3, BEST20.));
   run;
  
   %if &Param_1_init eq . OR Param_2_init eq . OR Param_3_init eq . %then %do;
      %let status = 1;
      %let reason = %sysfunc(sasmsg(&orvdds, lng_cant_compute_momest_error, noquote));
   %end;
   %else %do;
      ods listing;
      proc nlp  
              maxit  = 10000 
              maxfu  = 10000 
              tech   = tr 
              outest = _parmnlp_ 
              ;
         min obj;
         parms 
               Param_1 = &Param_1_init,
               Param_2 = &Param_2_init,
               Param_3 = &Param_3_init
               ;

         bounds %Param_BoundsLNG();
         %LNG_unboundedmoments;       
         f1 = (mean - &mean);
         f2 = (std *  std - &variance);
         f4 = (kurtosis - &kurtosis);
         obj = f1**2 + sqrt(f2**2) + (f4**2)**(.25);
      run; 
      quit;

      data _null_;
         set _parmnlp_;

         if _TYPE_ = "PARMS" then do;
            call symputx("Param_1", put(Param_1, BEST20.));
            call symputx("Param_2", put(Param_2, BEST20.));
            call symputx("Param_3", put(Param_3, BEST20.));
         end;
      run;  
      %if &SYSERR ^= 0 & &SYSERR ^= 4 %then %do;
         %put &orvar_error %sysfunc(sasmsg(&orvdds, lng_cant_find_est_error, noquote));
         %let status = 1;
         %let reason = %sysfunc(sasmsg(&orvdds, lng_cant_find_est_error, noquote));
      %end;
   %end;
%mend MomFit_real_LNG;

%macro Param_BoundsLNG();
   Param_2 > 0, 
   Param_3 > 1.0e-5
%mend;

%macro ValidateParamsLNG(Valid, Reason);
   if (Param_1 > 0) AND (Param_2 > 0) AND (Param_3 > 0) then do;
      &Valid  = "true";
   end;
   else do;
      &Valid  = "false";
      &Reason = sasmsg("&orvdds", "lng_one_param_not_pos_error", "quote");
   end;
%mend ValidateParamsLNG;

/*****************************************************************************************
 *   Needed to be called before anything else is available
 *****************************************************************************************/
%macro ORVD_LNG(minimalInit = true);   
   %orvd_logn(minimalInit = &minimalInit);

   %let NEEDED_INPUT    = ;
   %let SCALE_TRANSFORM = LOG;
   %let DIST_NAME       = Log Normal Gamma;
   %let PARAM_COUNT     = 3;
   %let LNG_NQUANT      = 31;
   %let ProbMeasure     = CONTINUOUS;

   %if %upcase(&minimalInit) NE TRUE %then %do;
      proc fcmp outlib = &ORVAR_CMPLIB..LNG;
         function Cdf_LNG(x,Param_1 &LNG_Param_List_Minus_1);
            if (Param_3) < 1e-8 then do;
               return( cdf("LOGNORMAL", x, Param_1, Param_2));   
            end;
            else do;
               tmp = 0;
               array q q_1-q_&LNG_NQUANT; 

               do i_lng_rep = 1 to &LNG_NQUANT;
                  if (Param_3) < 0.00002 then do;
                     q[i_lng_rep] = sqrt(1 + probit(i_lng_rep/(&LNG_NQUANT + 1)) * sqrt(Param_3));
                  end;
                  else do;
                     q[i_lng_rep] = sqrt(gaminv(i_lng_rep/(&LNG_NQUANT+1),1/(Param_3)) * (Param_3));
                  end;
               end;

               do i_lng_rep = 1 to &LNG_NQUANT;
                  if Param_2*q[i_lng_rep] < 1e-8 then do;
                     tmp = tmp + 1;
                  end;
                  else do;
                     tmp = tmp + cdf('LOGNORMAL', x, Param_1, Param_2*q[i_lng_rep]);
                  end; 
               end;

               return(tmp/&LNG_NQUANT);
            end;
         endsub;

         function Pdf_LNG(x,Param_1 &LNG_Param_List_Minus_1);
            if Param_3 < 1e-8 then do;
               return( pdf("LOGNORMAL", x, Param_1, Param_2));   
            end;
            else do;
               tmp = 0;
               array q q1-q&LNG_NQUANT; 

               do i_lng_rep = 1 to &LNG_NQUANT;
                  if Param_3 < 0.00002 then do;
                     q[i_lng_rep] = sqrt(1 + probit(i_lng_rep/(&LNG_NQUANT+1)) * sqrt(Param_3));
                  end;
                  else do;
                     q[i_lng_rep] = sqrt(gaminv(i_lng_rep/(&LNG_NQUANT+1), 1/Param_3) * Param_3);
                  end;
               end;

               do i_lng_rep = 1 to &LNG_NQUANT;
                  tmp = tmp + pdf('LOGNORMAL', x, Param_1, Param_2 * q[i_lng_rep]);
               end; 
               return(tmp);
            end;      
         endsub;
      run;
   %end;   
%DONE:
   data LNG_STATUS;
        status = 0;       
   run;
%mend ORVD_LNG;
