/*****************************************************************************************
 * Copyright (C), 2011
 * SAS Institute Inc., Cary, N.C. 27513, U.S.A.  All rights reserved.
 * 
 *  NAME:    orva_init.sas                                             
 *  PURPOSE: Performs initial set up for the OpRiskVaR application.
 *           This macro is called by the client shortly after connecting to the server.
 *           
 *  SUPPORT: jariou
 *  
 *  PARAMETERS: 
 *     DATA_ROOT - Path to datastore location to place datasets.  
 *                 Not required if resetOption = TEST
 *     resetOption - How to handle initialization files (distribution related datasets and 
 *             fcmp function libraries) that are stored in the datastore but accessed in work. 
 *              NONE - Most common case, just copy files from DATA_ROOT to WORK area
 *                      (if DATA_ROOT passed) and setup global macro variables.
 *                     When parallel processing is done, NONE will only set up global macro vars
 *              CREATE - recreate all files in DATA_ROOT (generally called by Config Editor), copy
 *                     to work area and set up global macro vars
 *              TEST - recreate all files directly in work area (for testing purposes) and 
 *                     set up global macro vars
 *     ctool - NOT USED!      
 *             
 *  NOTES:
 *     1. Parses Extensions.xml to create informational datasets:
 *             xml_lookup 
 *             sparmnames 
 *             fparmnames 
 *             scaletrans 
 *     2. Creates a set of global macro variables based on this information.
 *     3. Creates fcmp distortion functions.
 *     4. Creates fcmp distribution functions.
 *     5. Loads the distribution macros.
 *     6. Creates fcmp rare event simulation functions.
 *     7. Copies all this stuff to the work area
 *             
 *     Note that is resetOption=NONE only 2 and 7 are done. This is the
 *     most common use.
 *****************************************************************************************/
%macro orva_init( 
                 DATA_ROOT   = , 
                 ctool       = false, 
                 resetOption = TEST
                 );
   %if (%upcase(&ORV_DEBUG.) EQ TRUE) %then %do;
      %put #orva_init(; 
      %put .           DATA_ROOT   = &DATA_ROOT.,; 
      %put .           ctool       = &ctool.,; 
      %put .           resetOption = &resetOption.;
      %put .           );
      %put .;
   %end;

   /*** this global variable keeps this macro from running numerous times per session ***/
   %global ORVA_INITED;
   %global configLib;
     
   /*** only perform setup once per sas session ***/
   %if &ORVA_INITED ^= %str() %then %do;
     %let LOCAL_INITED_STATUS = 0;
     %goto DONE;
   %end;
   
   /*** in case a syntax error occurs, preset status to bad ***/
   %let LOCAL_INITED_STATUS = 1;
   data orvinited; 
      status = 1; 
   run;
  
   /*** setup destination libref (destlib) as requested ***/
   %let resetOption = %upcase(&resetOption); 

   %if "&DATA_ROOT" ne "" %then %do;
      libname destlib "&DATA_ROOT";
   %end;
   /* If we are in test mode, assign destlib to work */
   %else %if resetOption = TEST %then %do;
         libname destlib (work);
   %end;
   %else %if &resetOption eq CREATE %then %do;
      %put &orvar_error %sysfunc(sasmsg(&orvads, orva_init_error_parm1, noquote));
      %let LOCAL_INITED_STATUS = 1;
      %goto DONE;
   %end;
    
   /*** setup where to create datasets & function libraries based on resetOption ***/
   %if &resetOption eq CREATE %then %do;
     %let configLib = destlib;
     %orvu_delete_data( 
                        data = 
                               xml_lookup 
                               sparmnames 
                               sparmnames_2 
                               fparmnames 
                               fparmnames_2 
                               scaletrans 
                               orvfuncs,
                        lib   = destlib
                        );
     %_orv_CopyXmlFiles;
   %end;
   %else %do;
     %let configLib = work;
   %end;
   
   %let ORVAR_CMPLIB = &configLib..orvfuncs;
   
   /*-------------------------------------------------------------------------------- 
    * This case (NONE) happens in normal VaR client connects and parallel connects
    * 1. In normal VaR connects DATA_ROOT will be passed so we copy everything
    *    from DATA_ROOT to WORK
    * 2. In parallel connects, DATA_ROOT will not be passed because all
    *    the files will have been uploaded to WORK for us
    *-------------------------------------------------------------------------------*/
   %if &resetOption eq NONE %then %do; 
      %if "&DATA_ROOT" ne "" %then %do;
         %_orv_CopyDatasets(  
                            from = destlib, 
                            to   = work
                            );
      %end;

      libname destlib (work);
      %orva_init_global_macros(infolib = &configLib);
      %_orv_LoadDistMacros(
                           infolib     = &configLib, 
                           minimalInit = true
                           );
      %orvu_append_cmplib(
                          libname  = sashelp.svrtdist, 
                          position = back
                          );
      %orvu_append_cmplib(
                          libname  = &ORVAR_CMPLIB,    
                          position = front
                          );
   %end;
   /*-------------------------------------------------------------------------------- 
    * This case happens in sdstest (resetOption=TEST), proc anchor(resetOption=TEST), 
    * and when creating the datastore (resetOption=CREATE)
    * In all 3 cases we want to create all the files.
    * 1. In the case of creating the datastore DATA_ROOT will be passed so 
    *    everything will be created there and then copied to WORK.
    * 2. In the other 2 cases, everything will be created in WORK.
    *--------------------------------------------------------------------------------*/
   %else %do;
      %let HOST_MISC_ROOT =;
      %let HOST_CONFIG_MISC_ROOT =;

      /* Create some functions in Permlib  to suppress warnings */
      %orva_init_cont_empirical_func( funclib = &ORVAR_CMPLIB..EMP );
      %orva_init_disc_empirical_func( funclib = &ORVAR_CMPLIB..DEmp );
      %_orv_CreateBooleans( Lib = &configLib..Permlib );

      %_orv_ParseDistributions( outlib = &configLib );
      %orva_init_global_macros( infolib = &configLib );
      %orva_init_global_functions( cmplib  = &ORVAR_CMPLIB );
      %_orv_LoadDistMacros(
                           infolib     = &configLib, 
                           minimalInit = false
                           );
      %orvu_append_cmplib(
                          libname  = sashelp.svrtdist, 
                          position = back
                          );
      %orvu_append_cmplib(
                          libname  = &ORVAR_CMPLIB.,    
                          position = front
                          );

      %orva_init_distributions( 
                               freqNames = &configLib..fparmnames,
                               sevNames  = &configLib..sparmnames,
                               cmpLib    = &ORVAR_CMPLIB,
                               package   = dists,
                               paramName = ___x___
                               );    
      %_orv_BuildScaleTransform( 
                                sevConfigData    = &configLib..sparmnames_2, 
                                scaleTransformDS = &configLib..scaletrans 
                                );
      %orva_init_rare_scen_sim( 
                                 cmpLib   = &ORVAR_CMPLIB,
                                 package  = rare_sim 
                                 );

      %if &resetOption ne TEST and "&DATA_ROOT" ne "" %then %do;
         %_orv_CopyDatasets(  
                            from = destlib,
                            to   = work
                            );
      %end;
   %end;

   /* turn off warning for long strings ----------------------------------*/
   %if (%upcase(&ORV_DEBUG.) NE TRUE) %then %do;
      options NoQuoteLenMax;
   %end;
 
   /*** all is good now ***/
   %let LOCAL_INITED_STATUS = 0;
%DONE:
   /*** mark set-up complete ***/
   %let ORVA_INITED = 1;
   data orvinited; 
      status = &LOCAL_INITED_STATUS; 
   run;
%mend orva_init;

%macro _orv_CopyXmlFiles;
   %if %index(DNTHOST, &sysscp) %then %do;
     %let HOST_MISC_ROOT = %sysget(TESTsasroot)\orvar\misc;
     %let HOST_CONFIG_MISC_ROOT = %sysget(TESTsasroot)\orvar\misc;
   %end;
   %else %if %index(WIN, &sysscp) %then %do;
     %let HOST_MISC_ROOT = %sysget(SASROOT)\orvar\sasmisc;
     %let HOST_CONFIG_MISC_ROOT = %sysget(VARCONFIGROOT)\sasmisc;
   %end;
   %else %do;
     %let HOST_MISC_ROOT = %sysget(SASROOT)/misc/orvar;
     %let HOST_CONFIG_MISC_ROOT = %sysget(VARCONFIGROOT)/sasmisc;
   %end;
   
   %if %index(WIN DNTHOST, &sysscp) %then %do;
      options noxwait;
   %end;
   
   %orvu_file_copy(  
                     sourcedir = &HOST_CONFIG_MISC_ROOT,
                     source    = "Extensions.dtd",
                     destdir   = &DATA_ROOT, 
                     dest      = "Extensions.dtd",
                     encoding  = utf8
                     );
   %orvu_file_copy(  
                     sourcedir  = &HOST_CONFIG_MISC_ROOT,
                     source     = "Extensions.xml",
                     destdir    = &DATA_ROOT,
                     dest       = "Extensions.xml",
                     encoding   = utf8
                     );
   %orvu_file_copy(  
                     sourcedir  = &HOST_MISC_ROOT,
                     source     = "gde_historicalTrendStyle.xml",
                     destdir    = &DATA_ROOT,
                     dest       = "gdeHistoricalTrendStyle.xml",
                     encoding   = utf8
                     );
   %orvu_file_copy(  
                     sourcedir  = &HOST_MISC_ROOT,
                     source     = "gde_historicalTrendDetail.xml",
                     destdir    = &DATA_ROOT,
                     dest       = "gdeHistoricalTrendDetail.xml",
                     encoding   = utf8
                     );
   %orvu_file_copy(  
                     sourcedir  = &HOST_MISC_ROOT,
                     source     = "gde_historicalTrendOlap.xml",
                     destdir    = &DATA_ROOT,
                     dest       = "gdeHistoricalTrendOlap.xml",
                     encoding   = utf8
                     );
%mend _orv_CopyXmlFiles;

/**
 * Copies all datasets created in this macro from DATA_ROOT to WORK
 * WARNING: Don't change this list without updating %orvp_signon().
 */
%macro _orv_CopyDatasets(
                         from = , 
                         to   = 
                         );
   %if (%upcase(&ORV_DEBUG.) EQ TRUE) %then %do;
      %put @_orv_CopyDatasets(;
      %put .                   from = &from,;
      %put .                   to   = &to;
      %put .                   );
   %end;

   %put %sysfunc(sasmsg(&orvads, orva_init_copying_datasets, noquote, &from, &to));
   proc datasets nolist nodetails;
     copy 
            in  = &from 
            out = &to
            ;
     select 
            xml_lookup 
            sparmnames 
            sparmnames_2 
            fparmnames 
            fparmnames_2 
            scaletrans 
            orvfuncs
            permlib
            ;
   quit;
%mend _orv_CopyDatasets;

%macro _orv_ParseDistributions(outlib = );
   %if (%upcase(&ORV_DEBUG.) EQ TRUE) %then %do;
      %put @_orv_ParseDistributions(;
      %put .                         outlib = &outlib;
      %put .                         );
   %end;

   %let HOST_CONFIG_MISC_ROOT = ;

   %if %index(DNTHOST, &sysscp) %then %do;
      %let HOST_CONFIG_MISC_ROOT = %sysget(TESTsasroot)\orvar\misc;
   %end;
   %else %if %index(WIN, &sysscp) %then %do;
      %let HOST_CONFIG_MISC_ROOT = %sysget(VARCONFIGROOT)\sasmisc;
   %end;
   %else %do;
      %let HOST_CONFIG_MISC_ROOT = %sysget(VARCONFIGROOT)/sasmisc;
   %end;
   
   %let xmlFile = &HOST_CONFIG_MISC_ROOT/Extensions.xml;
   
   %orva_init_distortions( 
                           xmlFile = &xmlFile, 
                           outData = &outlib..distortions 
                           );
   %orvu_translate(
                     inFile  = SASHELP.orvxmsg, 
                     outFile = &outlib..xml_lookup
                     );
   %orvu_translate(
                     inFile       = SASHELP.orvomsg, 
                     outFile      = WORK.options_lookup,     
                     prefix       = FITOPTS_, 
                     filterString = _GUI
                     );
   %orvu_translate(
                     inFile       = SASHELP.orvamsg, 
                     outFile      = WORK.reason_lookup,      
                     prefix       = GRADREASON_, 
                     filterString = VARGRAD_
                     );
   %orvu_translate(
                     inFile       = SASHELP.orvamsg, 
                     outFile      = WORK.reason_coni_lookup, 
                     prefix       = GRADREASON_, 
                     filterString = CONI_
                     );

   data &outlib..xml_lookup;
      set 
            &outlib..xml_lookup 
            WORK.options_lookup
            ;
   run;
   
   proc append 
            base = &outlib..xml_lookup 
            data = WORK.reason_lookup
            ;
   run;
   
   proc append 
            base = &outlib..xml_lookup 
            data = WORK.reason_coni_lookup
            ;
   run;

   %orvu_delete_data( data = 
                              options_lookup 
                              reason_lookup 
                              reason_coni_lookup
                              );
   %orva_init_distribution_names(
                                 xmlFile = &xmlFile, 
                                 xmlMsgs = SASHELP.orvxmsg, 
                                 outData = &outlib..sparmnames, 
                                 type    = SEV
                                 );
   %orva_init_distribution_names(
                                 xmlFile = &xmlFile, 
                                 xmlMsgs = SASHELP.orvxmsg, 
                                 outData = &outlib..fparmnames, 
                                 type    = FREQ
                                 );
%mend _orv_ParseDistributions;

%macro _orv_BuildScaleTransform( 
                                sevConfigData    = , 
                                scaleTransformDS = 
                                );
   %if (%upcase(&ORV_DEBUG.) EQ TRUE) %then %do;
      %put @_orv_BuildScaleTransform(; 
      %put .                          sevConfigData    = &sevConfigData,; 
      %put .                          scaleTransformDS = &scaleTransformDS;
      %put .                          );
   %end;

   %let theSevDists = ;

   proc sql noprint;
      SELECT DISTINCT 
                     distribution
      INTO: 
                     theSevDists SEPARATED BY " "
      FROM 
                     &sevConfigData
                     ;
   quit;

   %let ndist = 1;
   %let dist  = %scan(&theSevDists, &ndist, ' '); 

   %do %while (&dist ne );  
      %orvd_&dist; 

      %if &nDist = 1 %then %do;
         data  &scaleTransformDS;
            length SCALE_TRANSFORM $5 Distribution $20;
            Distribution = "&dist";
            SCALE_TRANSFORM = "&SCALE_TRANSFORM";
         run;
      %end;
      %else %do;
         data  &scaleTransformDS;
            length 
                  SCALE_TRANSFORM $5 
                  Distribution $20
                  ;
            set &scaleTransformDS end = last;
            output;           
   
            if last then do;
               Distribution    = "&dist";
               SCALE_TRANSFORM = "&SCALE_TRANSFORM";
               output;
            end;
         run;
      %end;

      %let ndist = %eval(&ndist + 1);
      %let dist  = %scan(&theSevDists, &ndist, ' ');
   %end;
%mend _orv_BuildScaleTransform;

%macro _orv_LoadDistMacros(
                           infolib     = ,
                           minimalInit = 
                           );
   %if (%upcase(&ORV_DEBUG.) EQ TRUE) %then %do;
      %put @_orv_LoadDistMacros(;
      %put .                     infolib     = &infolib,;
      %put .                     minimalInit = &minimalInit;
      %put .                     );
   %end;

   %global theFreqSimCallParams;
   %global theSevSimCallParams; 

   /***************************************************************/
   /*** set macro var theFreqSimCallParams ************************/
   /***************************************************************/  
   proc sort 
         data = &infolib..fparmnames
         out  = _parmNames
         nodupkey
         ; 
      by DatasetName;
   run;

   proc SQL noprint;
      SELECT 
               DatasetName 
      INTO:  
               theFreqSimCallParams SEPARATED BY ", "
      FROM   
               _parmNames
               ;
   quit;  
   
   /***************************************************************/
   /*** load FREQ distributions ***********************************/
   /***************************************************************/
   %let theFreqDists = ;

   proc sql noprint;
      SELECT DISTINCT 
                  distribution
      INTO: 
                  theFreqDists SEPARATED BY " "
      FROM 
                  &infolib..fparmnames_2
                  ;
   quit;
 
   %let ndist = 1;
   %let dist = %scan(&theFreqDists, &ndist, ' '); 

   %do %while (&dist ne );     
      %orvd_&dist(minimalInit = &minimalInit); 

      proc sql noprint; 
         DROP TABLE 
                  &dist._status
                  ; 
      quit;

      %let ndist = %eval(&ndist + 1);
      %let dist  = %scan(&theFreqDists, &ndist, ' ');
   %end;
   
   /***************************************************************/
   /*** set macro var theSevSimCallParams ************************/
   /***************************************************************/  
   proc sort 
         data = &infolib..sparmnames
         out  = _parmNames
         nodupkey
         ; 
      by DatasetName;
   run;

   proc SQL noprint;
      SELECT 
               DatasetName 
      INTO:  
               theSevSimCallParams SEPARATED BY ", "
      FROM   
               _parmNames
               ;
   quit;
   
   /***************************************************************/
   /*** load SEVERITY distributions *******************************/
   /***************************************************************/
   %let theSevDists = ;

   proc sql noprint;
      SELECT DISTINCT 
                  distribution
      INTO: 
                  theSevDists SEPARATED BY " "
      FROM 
                  &infolib..sparmnames_2
                  ;
   quit;
   
   %let BODY_DIST = %str();
   %let outDist   = %str();
   %let ndist     = 1;
   %let dist = %scan(&theSevDists, &ndist, ' '); 

   %do %while (&dist ne );  
      %orva_set_sev_dist( distIn = &dist, distOut = OutDist );
      %orvd_&OutDist( minimalInit = &minimalInit ); 

      proc sql noprint; 
         DROP TABLE 
                  &OutDist._status
                  ;
      quit;

      %let ndist = %eval(&ndist + 1);
      %let dist = %scan(&theSevDists, &ndist, ' ');
   %end;
   
   proc sql noprint; 
      DROP TABLE 
               _parmNames
               ; 
   quit;
%mend _orv_LoadDistMacros;

%macro _orv_CreateBooleans(Lib =);
   proc fcmp outlib = &LIB..Booleans;
      function True();
         return(1);
      endsub;
      function False();
         return(0);
      endsub;
   run;
%mend _orv_CreateBooleans;

/******************************************************************/
/* Code to be run by orva_sfit_dist and orva_sfitkri_dist to save */
/* some time when the fitted datasets are very large              */
/******************************************************************/
%macro orv_optimize_fitting(withKRIs = NO);
   %put Optimizing fitting using random SUBSET .......;

   %local ORV_SAMPLE_SUBSET_SIZE;
   %let   ORV_SAMPLE_SUBSET_SIZE = 25000;

   /* No optimization for LNGROBUST */
   %if %upcase(&Dist.) EQ LNGROBUST or %upcase(&Dist.) EQ EMP %then %do;
      %goto GET_OUT_NOW;
   %end;

   /* Handle with or without KRI fitting */
   %if %upcase(&withKRIs. EQ NO) %then %do;
      %let FittingMacro = Fit&dist.;
      %let FittedData   = _logworkdata_;
   %end;
   %else %do;
      %let FittingMacro = FitKRI&dist.;
      %let FittedData   = _LogWorkdata_KRI_;
   %end;

   %let _orig_obsCount = &obsCount.;
   %let sampleProp     = %sysevalf(&ORV_SAMPLE_SUBSET_SIZE. / &obsCount);

   %if &sampleProp. LT 0.25 %then %do;
      data _orig_data_;
         set &FittedData.;
      run;

      data &FittedData.;
         set &FittedData. end = last;
         retain newCount 0;

         if ranuni(1234) LT &sampleProp. then do;
            newCount = newCount + 1;
            output;
         end;

         if last then do;
            call symput("obsCount", newCount);
         end;
      run;
      
      proc sort data = &FittedData.;
         by descending amount;
      run;

      %orvu_delete_data( data = _cs_ );
      %&FittingMacro.;

      %let PROC_NLP_OPTIONS = %str(&LAST_PROC_NLP_OPTIONS.);

      %if &status = 0 %then %do;
         %let nobs = 0;
         %let dsid = %sysfunc(open(_cs_, I));

         %if &dsid ^= 0 %then %do;
            %let nobs = %sysfunc(attrn(&dsid, nlobs));
            %let rc   = %sysfunc(close(&dsid));
         %end;

         %if &nobs eq 0 %then %do;
            %let status = 1;
            %let reason = %sysfunc(sasmsg(&orvads, sfdist_conv_fail_error, noquote));
         %end;
         %else %do;
            data _null_;
               set _cs_;

               if (substr(reason, 1, 7) eq "WARNING") then do;
                  status = 0;
               end;

               call symput("status", status);
               call symputX("reason", reason);
            run;
         %end;
      %end;

      %if &status = 0 %then %do;
         data _parmnlp_;
            set _parmnlp_ (where = (index(_type_, "PARMS") > 0));
         run;
         
         proc transpose 
                       data = _parmnlp_(drop = _NAME_)
                       out  = _param_vals_
                       ;
            var Param_:;
         run;


         data _NULL;
            set _param_vals_;
            where length(_NAME_) LT 10;

            if upcase(_NAME_) EQ "PARAM_0" then do;
               _NAME_ = "Param_1";
            end;

            put _NAME_ " = " Col1;
            call symput(trim(_NAME_) || "_init", put(col1, BEST20.));
         run;

         %if %upcase(&withKRIs. EQ YES) %then %do;
            PROC SQL noprint;
               SELECT
                     cat(_NAME_, " = ", put(col1, BEST20.))
               INTO
                     :KRI_INIT_PARM SEPARATED BY ","
               FROM
                     _param_vals_
               WHERE
                     upcase(substr(_NAME_, 1, 12)) EQ "PARAM_1_KRI_"
                     ;
            QUIT;
         
            %put KRI_INIT_PARM = &KRI_INIT_PARM.;
         %end;
      %end;

      %orvu_delete_data( data = 
                               _param_vals_ 
                               _parmnlp_
                               );

      /* Reset original parameters */
      %let obsCount = &_orig_obsCount.;

      data &FittedData.;
         set _orig_data_;
      run;

      %orvu_delete_data( data = _orig_data_ );
   %end;

%GET_OUT_NOW:
%mend orv_optimize_fitting;
     
